import numpy as np
import sys
import re
from ..coords import rotationAboutAxis

from ..segmentmodel import SegmentModel, SingleStrandedSegment, DoubleStrandedSegment

class MayaObj():
    def __init__(self, maya_lines):
        self.name, self.parent_name = MayaObj._parse_first_maya_line( maya_lines[0] )

    def _parse_first_maya_line(l):
        ## createNode vHelix -n "helix_1";
        name = parent = None
        vals = l.split()
        for i in range(len(vals)):
            if vals[i] == "-n": 
                name = MayaObj._strip_maya_string( vals[i+1] )
            if vals[i] == "-p":
                parent = MayaObj._strip_maya_string( vals[i+1] )
        return name,parent
    
    def _strip_maya_string(string):
        return re.match('"([a-zA-Z_[0-9]+)";?$', string).group(1)


class MayaConnection():
    def __init__(self, helix1, base1, suff1, helix2, base2, suff2):
        self.helix1 = helix1
        self.base1  = base1
        self.suff1  = suff1
        self.helix2 = helix2
        self.base2  = base2
        self.suff2  = suff2

# MayaConnections = dict()        

def ParseMayaConnection(line, base_dict):
    words = line.split()
    if len(words) != 3:
        return None
    cmd,obj1,obj2 = words

    def parse_obj(obj):
        ## Typically, base name is not unique, so parent name used too
        result = re.match('"(.*(forw)?(backw)?_[0-9]+)\.([a-z]{2})"', obj)
        if result is not None:
            name, tmp, tmp, suff = result.groups()
            if name in base_dict:
                return base_dict[name].parent.name, name, suff
            else:
                result = re.match('\|(.*)\|(.*)', name)
                if result is not None:
                    helix, name = result.groups()
                    return helix, name, suff

        return None, None, None
                    
    helix1,base1,suff1 = parse_obj(obj1)
    if helix1 is not None:
        helix2,base2,suff2 = parse_obj(obj2)
        if helix2 is not None:
            return MayaConnection(helix1,base1,suff1,helix2,base2,suff2)
    return None


class MayaHelix(MayaObj):
    def __init__(self, maya_lines):
        MayaObj.__init__(self, maya_lines)
        # self.bases = []
        self.bases = dict()
        self.position = None
        self.orientation = None

        self._parse_maya_lines(maya_lines)
        
    def _parse_maya_lines(self, maya_lines):
        """
createNode vHelix -n "helix_30";
setAttr ".t" -type "double3" 9.24809 21.4638 -12.465 ;
setAttr ".r" -type "double3" 28.99971393816978 2.2177410772603801 87.637950315471372 ;
setAttr ".dh" yes;
        """
        for l in maya_lines:
            vals = l.split()
            if vals[0] == "setAttr":
                if vals[1] == '".t"':
                    self.position = 10*np.array([float(a) for a in vals[4:7]])
                elif vals[1] == '".r"':
                    # self.eulerAngles = np.array([float(a) for a in vals[4:7]]) # defined here
                    ax,ay,az = [float(a) for a in vals[4:7]] 
                    Rx = rotationAboutAxis( [1,0,0], ax, normalizeAxis=False )
                    Ry = rotationAboutAxis( [0,1,0], ay, normalizeAxis=False )
                    Rz = rotationAboutAxis( [0,0,1], az, normalizeAxis=False )
                    self.orientation = Rz.dot(Ry.dot(Rx)) # TODO: check
                    # rotXYZ = np.array()

    def add_base(self, base):
        # self.bases.append(base)
        self.bases[base.name] = base
        base.parent = self

    def find_double_helix_regions(self):
        double_helices = []
        for bn,b1 in self.bases.items():
            b2 = b1.basepair
            if b2 is None: continue
            d =  np.linalg.norm( b1.get_position()-b2.get_position() )
            print( d )
            

    def write_xyz(self,file_handle = sys.stdout):
        for bn,b in self.bases.items():
            position = b.get_position()
            # position = b.position.dot( self.orientation ) + self.position # TODO: double check
            file_handle.write("%s %f %f %f\n" % (b.name[0], position[0], position[1], position[2]))


class MayaBase(MayaObj):
    def __init__(self, maya_lines):
        MayaObj.__init__(self, maya_lines)
        self._parse_maya_lines(maya_lines)        
        self.basepair = None
        self.parent = None
        self.end3 = None
        self.end5 = None

    def _parse_maya_lines(self, maya_lines):
        """
createNode HelixBase -n "backw_1" -p "helix_30";
setAttr ".t" -type "double3" 0.36767788771440857 -0.78848777472188547 -6.014 ;
        """
        for l in maya_lines:
            vals = l.split()
            if vals[0] == "setAttr":
                if vals[1] == '".t"':
                    self.position = 10*np.array([float(a) for a in vals[4:7]])

    def add_basepair(self, base):
        assert( self.parent == base.parent )
        assert( self.basepair is None or self.basepair == base )
        assert( base.basepair is None or base.basepair == self )
        self.basepair = base
        base.basepair = self

    def add_end3(self, base):
        assert( self.end3 is None or self.end3 == base )
        assert( base.end5 is None or base.end5 == self )
        self.end3 = base
        base.end5 = self

    def add_end5(self, base):
        base.add_end3(self)
    
    def get_position(self):
        h = self.parent
        return h.orientation.dot(self.position) + h.position # TODO: double check
        

# class MayaAimConstraint(MayaObj):
#     ...


# set ch [open /home/cmaffeo2/Downloads/ball.ma]


def parse_maya_file(maya_file):

    helices = dict()
    bases = []
    aimConstraints = []
    connections = []

    ## Parse through .ma file
    with open(maya_file) as fh:
        linesBuffer = []
        for line in fh:
            words = line.split()
            if words[0] == "createNode":

                ## Create a new object
                if len(linesBuffer) > 0:
                    objType = linesBuffer[0].split()[1]
                    if objType == "vHelix":
                        h = MayaHelix( linesBuffer )
                        helices[h.name] = h
                    elif objType == "HelixBase":
                        bases.append( MayaBase( linesBuffer ) )
                    # elif objType == "aimConstraint":
                    #     aimConstraints.append( MayaAimConstraint( linesBuffer ) )
                ## Clear lines buffer
                linesBuffer = []
            elif words[0] == "connectAttr":
                connections.append(line)
                
            ## Extend lines buffer
            linesBuffer.append(line)
            
    ## Crawl through bases, attach to parent helix
    for b in bases:
        helix_name = b.parent_name
        helices[helix_name].add_base( b ) # function updates base position?

    ## Parse connections
    new_connections = []
    base_dict = { b.name:b for b in bases }
    for line in connections:        
        conn = ParseMayaConnection(line, base_dict)
        if conn is not None:
            new_connections.append(conn)
    connections = new_connections

    ## Assign base connectivity
    for c in connections:
        h1 = c.helix1
        bn1 = c.base1
        s1 = c.suff1
        h2 = c.helix2
        bn2 = c.base2
        s2 = c.suff2
        
        b1 = helices[h1].bases[bn1]
        b2 = helices[h2].bases[bn2]
        if s1 == "lb" and s2 == "lb":
            b1.add_basepair(b2)
        # elif (bn1[0] == "f" and s1 == "fw" and s2 == "bw") or \
        #      (bn1[0] == "b" and s1 == "bw" and s2 == "fw"):
        #     b1.add_end3(b2)
        # elif (bn1[0] == "f" and s1 == "bw" and s2 == "fw") or \
        #      (bn1[0] == "b" and s1 == "fw" and s2 == "bw"):
        #     b1.add_end5(b2)
        elif s1 == "fw" and s2 == "bw":
            b1.add_end3(b2)
        elif s1 == "bw" and s2 == "fw":
            b1.add_end5(b2)
        else:
            raise Exception("Unkown connection %s %s" % (s1,s2))

    # # base_positions = np.zeros( [ len(bases), 3 ] )
    # base_positions = np.array( [b.position for b in bases] )
    # for aim in aimConstraints:
    #     base = base_dict[ aim.parent_name ]

    return helices

def convert_maya_to_segments(helices):

    def get_end3_ssDNA(nt):
        ret = []
        if nt.end3 is not None:
            # print( nt.parent.name, nt.end3.parent.name, nt.end3.name )
            nt2 = nt.end3
            if nt2.basepair is None:
                ret.append(nt2)
                ret.extend(get_end3_ssDNA(nt2))
        return ret

    def get_end5_ssDNA(nt):
        ret = []
        if nt.end5 is not None:
            nt2 = nt.end5
            if nt2.basepair is None:
                ret.append(nt2)
                ret.extend(get_end5_ssDNA(nt2))
        return ret

    """ Build dsDNA helices """
    segments = dict()
    segmentsBot = dict()
    segmentsTop = dict()
    for hn,h in helices.items():
        paired_bases = [b for bn,b in h.bases.items() if b.basepair is not None]
        upaired = [b for bn,b in h.bases.items() if b.basepair is None]
        # for b in upaired:
        #     print(hn, b.name, b.end3, b.end5)
        
        ## Find helical axis
        # if len(paired_bases) < 6:
        #     raise Exception("Helices shorter than 3 bp are not supported")
        # coords = np.array([b.position for b in paired_bases])
        # u,s,vh = np.linalg.svd(coords)
        # axis = vh[0] / np.linalg.norm(vh[0])

        fwd_bases = list(filter(lambda b:b.name[0] == 'f', paired_bases))
        # rank = [np.dot(b.position,axis) for b in fwd_bases]
        rank = [b.position[2] for b in fwd_bases]
        order = np.argsort(rank)
        ordered_fwd_bases = [fwd_bases[o] for o in order]
        next_fwd_is_3end = [b1.end3==b2 for b1,b2 in
                            zip(ordered_fwd_bases[:-1],ordered_fwd_bases[1:])]

        ## Check if axis points the wrong way
        assert( np.mean(next_fwd_is_3end) > 0.5 )
        # if np.mean( next_fwd_is_3end ) < 0.5:
        #     # axis = axis * -1
        #     rank = rank[order[-1]]-rank
        #     order = reversed(order)
        #     ordered_fwd_bases = reversed(ordered_fwd_bases)

        top = ordered_fwd_bases[-1]
        bot = ordered_fwd_bases[0].basepair

        assert( bot.name[0] == 'b' )
        assert( top.name[0] == 'f' )

        botPos = 0.5 * (bot.get_position() + bot.basepair.get_position())
        topPos = 0.5 * (top.get_position() + top.basepair.get_position())

        segname = hn
        if segname[:6] == "helix_":
            segname = 'D'+segname[6:]
        segments[hn] =  DoubleStrandedSegment(segname, num_bp = len(fwd_bases),
                                              start_position = botPos,
                                              end_position = topPos)

        ## Add 3prime/5prime ends
        nt = 0
        def add_primes(b,nt,fwd):
            if b.end3 is None:
                segments[hn].add_3prime(nt,fwd)
            if b.end5 is None:
                segments[hn].add_5prime(nt,fwd)

        for b in ordered_fwd_bases:
            add_primes(b,nt,True)
            add_primes(b.basepair,nt,False)
            nt+=1

        segmentsBot[hn] = bot
        segmentsTop[hn] = top
        assert( bot.end3 not in paired_bases )
        assert( top.end3 not in paired_bases )

    """ Sanity check """
    def end_crosses(end,direction):
        assert(direction in ("end3","end5"))
        other = end.__dict__[direction]
        if other is None or end.parent != other.parent:
            return 1
        else:
            return 0

    hs = segments.keys()
    botCrossings1 = [end_crosses(segmentsBot[h],"end3") for h in hs]
    botCrossings2 = [end_crosses(segmentsBot[h].basepair,"end5") for h in hs]
    topCrossings1 = [end_crosses(segmentsTop[h],"end3") for h in hs]
    topCrossings2 = [end_crosses(segmentsTop[h].basepair,"end5") for h in hs]
    # print( [np.mean(a) for a in (botCrossings1,botCrossings2,topCrossings1,topCrossings2)] )

    # for a in (botCrossings1,botCrossings2,topCrossings1,topCrossings2:
    #     print(np.mean(a))
    #     assert( np.mean(a) > 0.5 )

    """ Build ssDNA segments """
    ss_segments = []
    connectionToSingleStrand = dict()
    for hn,seg in segments.items():
        bot = segmentsBot[hn]
        top = segmentsTop[hn]
        # if top.end3 is not None:
        #     print(top.parent.name,top.name, top.end3.parent.name, top.end3.name)
        # if bot.end3 is not None:
        #     print(bot.parent.name,bot.name, bot.end3.parent.name, bot.end3.name)

        ## TODO: clean this up
        def get_end_positions(ssSeg, is_fwd):
            start_idx, end_idx = (0,-1)
            if not is_fwd:
                start_idx, end_idx = (-1,0)

            r0 = ssSeg[0].get_position()
            if ssSeg[0].end5 is not None:
                r0 = 0.75*r0 + 0.25*ssSeg[0].end5.get_position()
            r1 = ssSeg[-1].get_position()
            if ssSeg[-1].end3 is not None:
                r1 = 0.75*r1 + 0.25*ssSeg[-1].end3.get_position()
            return r0,r1

        if top not in connectionToSingleStrand:
            ssSeg = get_end3_ssDNA(top)
            num_nt = len(ssSeg)
            if num_nt > 0:
                r0,r1 = get_end_positions(ssSeg,is_fwd=True)
                seg2 = SingleStrandedSegment("S{:03d}".format(len(ss_segments)), 
                                              start_position = r0,
                                              end_position = r1,
                                              num_nt = num_nt)
                seg.connect_end3(seg2)
                connectionToSingleStrand[top] = seg2
                other = ssSeg[-1].end3
                if other is not None:
                    hn3 = other.parent.name
                    seg3 = segments[hn3]
                    bot3 = segmentsBot[hn3]
                    top3 = segmentsTop[hn3]
                    # if bot3 == other or bot3.basepair == other:
                    #     seg3.connect_start5(seg2)
                    # elif top3 == other or top3.basepair == other:
                    #     seg3.connect_end5(seg2)
                    if bot3.basepair == other:
                        seg3.connect_start5(seg2)
                    elif top3.basepair == other:
                        seg3.connect_end5(seg2)
                    else:
                        raise Exception("BUG")
                    connectionToSingleStrand[other] = seg2
                ss_segments.append(seg2)

        if top.basepair not in connectionToSingleStrand:
            ssSeg = get_end5_ssDNA(top.basepair)[::-1]
            if len(ssSeg) > 0:
                r0,r1 = get_end_positions(ssSeg,is_fwd=False)
                seg2 = SingleStrandedSegment("S{:03d}".format(len(ss_segments)),
                                              start_position = r0,
                                              end_position = r1,
                                              num_nt = len(ssSeg))
                seg.connect_end5(seg2)
                connectionToSingleStrand[top.basepair] = seg2
                other = ssSeg[0].end5
                if other is not None:
                    hn3 = other.parent.name
                    seg3 = segments[hn3]
                    bot3 = segmentsBot[hn3]
                    top3 = segmentsTop[hn3]
                    # if bot3 == other or bot3.basepair == other:
                    #     seg3.connect_start5(seg2)
                    # elif top3 == other or top3.basepair == other:
                    #     seg3.connect_end5(seg2)
                    if bot3 == other:
                        seg3.connect_start3(seg2)
                    elif top3 == other:
                        seg3.connect_end3(seg2)
                    else:
                        raise Exception("BUG")
                    connectionToSingleStrand[other] = seg2
                ss_segments.append(seg2)

        if bot not in connectionToSingleStrand:
            ssSeg = get_end3_ssDNA(bot)
            if len(ssSeg) > 0:
                r0,r1 = get_end_positions(ssSeg,is_fwd=True)
                seg2 = SingleStrandedSegment("S{:03d}".format(len(ss_segments)),
                                              start_position = r0,
                                              end_position = r1,
                                              num_nt = len(ssSeg))
                seg.connect_start3(seg2)
                connectionToSingleStrand[bot] = seg2
                other = ssSeg[-1].end3
                if other is not None:
                    hn3 = other.parent.name
                    seg3 = segments[hn3]
                    bot3 = segmentsBot[hn3]
                    top3 = segmentsTop[hn3]
                    if bot3.basepair == other:
                        seg3.connect_start5(seg2)
                    elif top3.basepair == other:
                        seg3.connect_end5(seg2)
                    else:
                        raise Exception("BUG")
                    connectionToSingleStrand[other] = seg2
                ss_segments.append(seg2)

        if bot.basepair not in connectionToSingleStrand:
            ssSeg = get_end5_ssDNA(bot.basepair)[::-1]
            if len(ssSeg) > 0:
                r0,r1 = get_end_positions(ssSeg,is_fwd=False)
                seg2 = SingleStrandedSegment("S{:03d}".format(len(ss_segments)),
                                              start_position = r0,
                                              end_position = r1,
                                              num_nt = len(ssSeg))
                seg.connect_start5(seg2)
                connectionToSingleStrand[bot.basepair] = seg2
                other = ssSeg[0].end5
                if other is not None:
                    hn3 = other.parent.name
                    seg3 = segments[hn3]
                    bot3 = segmentsBot[hn3]
                    top3 = segmentsTop[hn3]
                    if bot3 == other:
                        seg3.connect_start3(seg2)
                    elif top3 == other:
                        seg3.connect_end3(seg2)
                    else:
                        raise Exception("BUG")
                    connectionToSingleStrand[other] = seg2
                ss_segments.append(seg2)
                
    ## Connect remaining dsDNA ends
    no_connection_count = 0
    for hn,seg in segments.items():
        bot = segmentsBot[hn]
        top = segmentsTop[hn]

        top_bp, bot_bp = [b.basepair for b in (top,bot)]
        
        if top not in connectionToSingleStrand and top.end3 is not None:
            b2 = top.end3
            h2 = b2.parent.name
            s2 = segments[h2]
            if b2 == segmentsBot[h2] or b2.basepair == segmentsBot[h2]:
                seg.connect_end3(s2.start5,type_="terminal_crossover")
            elif b2 == segmentsTop[h2] or b2.basepair == segmentsTop[h2]:
                seg.connect_end3(s2.end5,type_="terminal_crossover")
            else:
                raise Exception("BUG")
        
        if top_bp not in connectionToSingleStrand and top_bp.end5 is not None:
            b2 = top_bp.end5
            h2 = b2.parent.name
            s2 = segments[h2]
            if b2 == segmentsBot[h2] or b2.basepair == segmentsBot[h2]:
                seg.connect_end5(s2.start3,type_="terminal_crossover")
            elif b2 == segmentsTop[h2] or b2.basepair == segmentsTop[h2]:
                seg.connect_end5(s2.end3, type_="terminal_crossover")
            else:
                raise Exception("BUG")

        if bot not in connectionToSingleStrand and bot.end3 is not None:
            b2 = bot.end3
            h2 = b2.parent.name
            s2 = segments[h2]
            if b2 == segmentsBot[h2] or b2.basepair == segmentsBot[h2]:
                seg.connect_start3(s2.start5,type_="terminal_crossover")
            elif b2 == segmentsTop[h2] or b2.basepair == segmentsTop[h2]:
                seg.connect_start3(s2.end5,type_="terminal_crossover")
            else:
                raise Exception("BUG")
        
        if bot_bp not in connectionToSingleStrand and bot_bp.end5 is not None:
            b2 = bot_bp.end5
            h2 = b2.parent.name
            s2 = segments[h2]
            if b2 == segmentsBot[h2] or b2.basepair == segmentsBot[h2]:
                seg.connect_start5(s2.start3,type_="terminal_crossover")
            elif b2 == segmentsTop[h2] or b2.basepair == segmentsTop[h2]:
                seg.connect_start5(s2.end3, type_="terminal_crossover")
            else:
                raise Exception("BUG")

    return [s for n,s in segments.items()] + ss_segments, segments


def write_xyz(filename, helices):
    nbeads = 0
    for hn,h in helices.items():
        nbeads += len(h.bases)

    with open("test.xyz","w") as fh:
        fh.write("%d\n\n" % nbeads)
        for hn,h in helices.items():
            h.write_xyz(fh)
            #      h.find_double_stranded_regions


if __name__ == "__main__":
    maya_helices = parse_maya_file("/home/cmaffeo2/Downloads/bunny.ma")

    segments,dsSegmentDict = convert_maya_to_segments(maya_helices)
    
    model = SegmentModel( segments,
                          local_twist = True,
                          max_basepairs_per_bead = 1,
                          max_nucleotides_per_bead = 1,
                          dimensions=(5000,5000,5000), DEBUG=True )

    model.simulate( outputPrefix = 'bunny2', outputPeriod=1e4, numSteps=1e8, gpu=1 )

